"""
Add @beautify behavior to enhance configuraptor/TypedConfig classes automagically!
"""

import functools
import typing

from .dump import asdict, asjson

T = typing.TypeVar("T")


def is_default(obj: typing.Any, prop: str) -> bool:
    """
    Check if the property of an object is set to its default value.

    Args:
        obj (typing.Any): The object to check.
        prop (str): The property to check.

    Returns:
        bool: True if the property is set to its default value, False otherwise.
    """
    return getattr(obj, prop) is getattr(object, prop)


def patch(cls: typing.Type[T], patch_repr: bool, patch_str: bool) -> None:
    """
    Patch the __str__ and __repr__ methods of a class if they are set to their default values.

    Args:
        cls (typing.Type[typing.Any]): The class to patch.
        patch_repr: patch __repr__? (if no custom one set yet)
        patch_str: patch __str__? (if no custom one set yet)
    """

    def _repr(self: T) -> str:
        """
        Custom __repr__ by configuraptor @beautify.
        """
        clsname = type(self).__name__
        data = asdict(self, with_top_level_key=False, exclude_internals=2)
        return f"<{clsname} {data}>"

    def _str(self: T) -> str:
        """
        Custom __str__ by configuraptor @beautify.
        """
        return asjson(self, with_top_level_key=False, exclude_internals=2)

    # if magic method is already set, don't overwrite it!
    if patch_str and is_default(cls, "__str__"):
        cls.__str__ = _str  # type: ignore

    if patch_repr and is_default(cls, "__repr__"):
        cls.__repr__ = _repr  # type: ignore


@typing.overload
def beautify(
    maybe_cls: typing.Type[T],
    repr: bool = True,  # noqa A002
    str: bool = True,  # noqa A002
) -> typing.Type[T]:
    """
    Overload function for the beautify decorator when used without parentheses.
    """


@typing.overload
def beautify(
    maybe_cls: None = None,
    repr: bool = True,  # noqa A002
    str: bool = True,  # noqa A002
) -> typing.Callable[[typing.Type[T]], typing.Type[T]]:
    """
    Overload function for the beautify decorator when used with parentheses.
    """


def beautify(
    maybe_cls: typing.Type[T] | None = None,
    repr: bool = True,  # noqa A002
    str: bool = True,  # noqa A002
) -> typing.Type[T] | typing.Callable[[typing.Type[T]], typing.Type[T]]:
    """
    The beautify decorator. Enhances a class by patching its __str__ and __repr__ methods.

    Args:
        maybe_cls (typing.Type[T] | None, optional): The class to beautify. None when used with parentheses.
        repr: patch __repr__? (if no custom one set yet)
        str: patch __str__? (if no custom one set yet)

    Returns:
        The beautified class or the beautify decorator.
    """
    if maybe_cls:
        patch(maybe_cls, patch_repr=repr, patch_str=str)
        return maybe_cls
    else:
        return functools.partial(beautify, repr=repr, str=str)



"""
File contains logic to do with the 'postpone' feature.
"""

from typing import Any, Optional

from typing_extensions import Never

from .errors import IsPostponedError
from .singleton import Singleton


class Postponed(Singleton):
    """
    Class returned by `postpone` below.
    """

    def __get_property_name__(self, instance: type[Any], owner: type[Any]) -> Optional[str]:
        """
        Internal method to get the property name of the class this descriptor is being used on.
        """
        if not instance:  # pragma: no cover
            return None

        # instance: the instance the descriptor is accessed from
        # owner: the class that owns the descriptor
        property_name = None
        for attr_name, attr_value in owner.__dict__.items():
            if attr_value is self:
                property_name = attr_name
                break
        # return instance.__dict__.get(property_name, None)
        return property_name

    def __get__(self, instance: type[Any], owner: type[Any]) -> Never:
        """
        This magic method is called when a property is accessed.

        Example:
             someclass.someprop will trigger the __get__ of `someprop`

        Args:
            instance: the class on which the postponed property is defined,
            owner: `SingletonMeta`
        """
        msg = f"Err: Using postponed property on {owner.__name__}"

        if name := self.__get_property_name__(instance, owner):
            msg += f".{name}"

        raise IsPostponedError(msg)


def postpone() -> Any:
    """
    Can be used to mark a property as postponed, meaning the user will fill it in later (they promose).

    If they don't fill it in and still try to use it, they will be met with a IsPostponedError.
    """
    return Postponed()



import types

from src import configuraptor


@configuraptor.converter(str, list[str])
def using_csv(data: str) -> list[str]:
    """
    Register a converter that will use values that are `str` but annotated as `list[str]` and convert by splitting.
    """
    return data.split(",")


@configuraptor.converter(str, list[int])
def list_of_numbers(data: str) -> list[int]:
    """
    Another example, will convert data to a list of ints, if the property is annotated as such.
    """
    list_of_string = using_csv(data)
    return [int(_) for _ in list_of_string]


class MyConfig:
    data: list[str]
    boolean: bool
    none: None
    null: types.NoneType
    number: float

    numbers: list[int]
    strings: list[str]


def test_csv_converter():
    config = configuraptor.load_into(
        MyConfig,
        {
            # custom:
            "data": "this,is,csv",
            "numbers": "1,22,3",
            "strings": "1,22,3",  # same data, different annotation -> different output
            # builtin converters:
            "boolean": "false",
            "none": "",
            "null": "",
            # fallback:
            "number": "3.14",
        },
        convert_types=True,
    )

    _this, _is, _csv = config.data

    assert _this == "this"
    assert _is == "is"
    assert _csv == "csv"

    assert config.boolean is False
    assert config.none is None
    assert config.null is None

    assert config.number == 3.14

    assert config.numbers == [1, 22, 3]
    assert config.strings == ["1", "22", "3"]



from configuraptor import Singleton, TypedConfig, load_into, postpone


class Later(TypedConfig, Singleton):
    field: str  # instant
    other_field: str = postpone()

    def update(self):
        self.other_field = "usable"


config = load_into(
    Later,
    {
        "later": dict(field="instant")
        # no other_field yet!
    },
)

print(config.field)  # will work
# config.other_field  # will give an error if you try to use it here!

config.update()  # fill in other_field some way or another
print(config.other_field)  # works now!



"""
Alias functionality so config keys can have multiple names.
"""

import typing
from dataclasses import dataclass
from typing import Any

from .abs import AnyType, T


@dataclass(frozen=True, slots=True)
class Alias:
    """
    Internal class used to relate keys.
    """

    to: str


def alias(to: str) -> Any:
    """
    Function to create an alias to a different key in the same class.
    """
    return Alias(to)


def has_aliases(cls: AnyType, key: str) -> typing.Generator[str, None, None]:
    """
    Generate all aliases that point to 'key' in 'cls'.
    """
    for field, value in cls.__dict__.items():
        if isinstance(value, Alias) and value.to == key:
            yield field


def has_alias(cls: AnyType, key: str, data: dict[str, T]) -> typing.Optional[T]:
    """
    Get the value of any alias in the same config class that references `key`.

    Example:
        class Config:
            key1: str
            key2: str = alias('key1')

    load_into(Config, {'key2': 'something'})
    # -> key1 will look up the value of key2 because it's configured as an alias for it.

    If multiple aliases point to the same base, they are all iterated until a valid value was found.
    """
    # for field, value in cls.__dict__.items():
    #     if isinstance(value, Alias) and value.to == key:
    #         # yay!
    #         return data.get(field)
    #
    # return None

    return next(
        (value for field in has_aliases(cls, key) if (value := data.get(field))),
        None,
    )


def is_alias(cls: AnyType, prop: str) -> bool:
    """
    Returns whether 'prop' is an alias to something else on cls.
    """
    return isinstance(cls.__dict__.get(prop), Alias)



"""
Logic for the TypedConfig inheritable class.
"""

import copy
import os
import typing
from collections.abc import Mapping, MutableMapping
from typing import Any, Iterator

from typing_extensions import Never, Self

from . import Alias
from .abs import AbstractTypedConfig
from .alias import has_aliases
from .beautify import beautify as apply_beautify
from .core import check_and_convert_type
from .errors import ConfigErrorExtraKey, ConfigErrorImmutable
from .helpers import all_annotations, is_optional
from .loaders.loaders_shared import _convert_key

C = typing.TypeVar("C", bound=Any)

NO_ANNOTATION = typing.NewType("NO_ANNOTATION", object)  # SentinelObject


class TypedConfig(AbstractTypedConfig):
    """
    Can be used instead of load_into.
    """

    def __init_subclass__(cls, beautify: bool = True, **_: typing.Any) -> None:
        """
        When inheriting from TypedConfig, automatically beautify the class.

        To disable this behavior:
        class MyConfig(TypedConfig, beautify=False):
            ...
        """
        if beautify:
            apply_beautify(cls)

    def _update(
        self,
        _strict: bool = True,
        _allow_none: bool = False,
        _skip_none: bool = False,
        _overwrite: bool = True,
        _ignore_extra: bool = False,
        _lower_keys: bool = False,
        _normalize_keys: bool = True,
        _convert_types: bool = False,
        _update_aliases: bool = True,
        **values: Any,
    ) -> Self:
        """
        Underscore version can be used if .update is overwritten with another value in the config.
        """
        annotations = all_annotations(self.__class__)

        for key, value in values.items():
            if _lower_keys:
                key = key.lower()

            if _normalize_keys:
                # replace - with _
                key = _convert_key(key)

            annotation = annotations.get(key, NO_ANNOTATION)

            if value is None and ((not is_optional(annotation) and not _allow_none) or _skip_none):
                continue

            existing_value = self.__dict__.get(key)
            if existing_value is not None and not _overwrite:
                # fill mode, don't overwrite
                continue

            if _strict and annotation is NO_ANNOTATION:
                if _ignore_extra:
                    continue
                else:
                    raise ConfigErrorExtraKey(cls=self.__class__, key=key, value=value)

            # check_and_convert_type
            if _strict and not (value is None and _allow_none):
                value = check_and_convert_type(value, annotation, convert_types=_convert_types, key=key)

            self.__dict__[key] = value
            # setattr(self, key, value)

            if _update_aliases:
                cls = self.__class__
                prop = cls.__dict__.get(key)
                if isinstance(prop, Alias):
                    self.__dict__[prop.to] = value
                else:
                    for alias in has_aliases(cls, key):
                        self.__dict__[alias] = value

        return self

    def update(
        self,
        _strict: bool = True,
        _allow_none: bool = False,
        _skip_none: bool = False,
        _overwrite: bool = True,
        _ignore_extra: bool = False,
        _lower_keys: bool = False,
        _normalize_keys: bool = True,
        _convert_types: bool = False,
        _update_aliases: bool = True,
        **values: Any,
    ) -> Self:
        """
        Update values on this config.

        Args:
            _strict: allow wrong types?
            _allow_none: allow None or skip those entries for required items?
            _skip_none: skip none also for optional items?
            _overwrite: also update not-None values?
            _ignore_extra: skip additional keys that aren't in the object.
            _lower_keys: set the keys to lowercase (useful for env)
            _normalize_keys: change - to _
            _convert_types: try to convert variables to the right type if they aren't yet
            _update_aliases: also update related fields?

            **values: key: value pairs in the right types to update.
        """
        return self._update(
            _strict=_strict,
            _allow_none=_allow_none,
            _skip_none=_skip_none,
            _overwrite=_overwrite,
            _ignore_extra=_ignore_extra,
            _lower_keys=_lower_keys,
            _normalize_keys=_normalize_keys,
            _convert_types=_convert_types,
            _update_aliases=_update_aliases,
            **values,
        )

    def __or__(self, other: dict[str, Any]) -> Self:
        """
        Allows config |= {}.

        Where {} is a dict of new data and optionally settings (starting with _)

        Returns an updated clone of the original object, so this works too:
        new_config = config | {...}
        """
        to_update = self._clone()
        return to_update._update(**other)

    def update_from_env(self) -> Self:
        """
        Update (in place) using the current environment variables, lowered etc.

        Ignores extra env vars.
        """
        return self._update(_ignore_extra=True, _lower_keys=True, _convert_types=True, **os.environ)

    def _fill(self, _strict: bool = True, **values: typing.Any) -> Self:
        """
        Alias for update without overwrite.

        Underscore version can be used if .fill is overwritten with another value in the config.
        """
        return self._update(_strict, _allow_none=False, _overwrite=False, **values)

    def fill(self, _strict: bool = True, **values: typing.Any) -> Self:
        """
        Alias for update without overwrite.
        """
        return self._update(_strict, _allow_none=False, _overwrite=False, **values)

    @classmethod
    def _all_annotations(cls) -> dict[str, type]:
        """
        Shortcut to get all annotations.
        """
        return all_annotations(cls)

    def _format(self, string: str) -> str:
        """
        Format the config data into a string template.

        Replacement for string.format(**config), which is only possible for MutableMappings.
        MutableMapping does not work well with our Singleton Metaclass.
        """
        return string.format(**self.__dict__)

    def __setattr__(self, key: str, value: typing.Any) -> None:
        """
        Updates should have the right type.

        If you want a non-strict option, use _update(strict=False).
        """
        if key.startswith("_"):
            return super().__setattr__(key, value)
        self._update(**{key: value})

    def _clone(self) -> Self:
        return copy.deepcopy(self)


K = typing.TypeVar("K", bound=str)
V = typing.TypeVar("V", bound=Any)


class TypedMappingAbstract(TypedConfig, Mapping[K, V]):
    """
    Note: this can't be used as a singleton!

    Don't use directly, choose either TypedMapping (immutable) or TypedMutableMapping (mutable).
    """

    def __getitem__(self, key: K) -> V:
        """
        Dict-notation to get attribute.

        Example:
            my_config[key]
        """
        return typing.cast(V, self.__dict__[key])

    def __len__(self) -> int:
        """
        Required for Mapping.
        """
        return len(self.__dict__)

    def __iter__(self) -> Iterator[K]:
        """
        Required for Mapping.
        """
        # keys is actually a `dict_keys` but mypy doesn't need to know that
        keys = typing.cast(list[K], self.__dict__.keys())
        return iter(keys)


class TypedMapping(TypedMappingAbstract[K, V]):
    """
    Note: this can't be used as a singleton!
    """

    def _update(self, *_: Any, **__: Any) -> Never:
        raise ConfigErrorImmutable(self.__class__)


class TypedMutableMapping(TypedMappingAbstract[K, V], MutableMapping[K, V]):
    """
    Note: this can't be used as a singleton!
    """

    def __setitem__(self, key: str, value: V) -> None:
        """
        Dict notation to set attribute.

        Example:
            my_config[key] = value
        """
        self.update(**{key: value})

    def __delitem__(self, key: K) -> None:
        """
        Dict notation to delete attribute.

        Example:
            del my_config[key]
        """
        del self.__dict__[key]

    def update(self, *args: Any, **kwargs: V) -> Self:  # type: ignore
        """
        Ensure TypedConfig.update is used en not MutableMapping.update.
        """
        return TypedConfig._update(self, *args, **kwargs)


T = typing.TypeVar("T", bound=TypedConfig)


# also expose as separate function:
def update(self: T, _strict: bool = True, _allow_none: bool = False, **values: Any) -> T:
    """
    Update values on a config.

    Args:
        self: config instance to update
        _strict: allow wrong types?
        _allow_none: allow None or skip those entries?
        **values: key: value pairs in the right types to update.
    """
    return TypedConfig._update(self, _strict, _allow_none, **values)



import json
import struct

import tomli_w
import yaml

from src import configuraptor
from src.configuraptor import BinaryConfig, BinaryField, asbytes


class MyBinaryConfig(BinaryConfig):
    # annotations not supported! (because mixing annotation and __dict__ lookup messes with the order,
    # which is important for struct.(un)pack
    number = BinaryField(int)
    string = BinaryField(str, length=5)
    decimal = BinaryField(float)
    double = BinaryField(float, format="d")
    other_string = BinaryField(str, format="10s")
    boolean = BinaryField(bool)


class TopLevel:
    binary: MyBinaryConfig


class JsonField:
    name: str
    age: int


class NestedBinaryConfig(BinaryConfig):
    data1 = BinaryField(JsonField, format="json", length=32)
    data2 = BinaryField(JsonField, format="yaml", length=32)
    data3 = BinaryField(JsonField, format="toml", length=32)
    other_data = "don't touch this"


def test_binary_config():
    binary = struct.pack("I 5s f d 10s b", 42, "Hello".encode(), 3.6, 10 / 3, b"Hi", True)
    data = {"binary": binary}

    inst = configuraptor.load_into(TopLevel, data).binary
    assert MyBinaryConfig.load(binary)

    assert inst.string == "Hello"
    assert inst.other_string == "Hi"
    assert inst.boolean is True

    assert inst._pack() == binary == asbytes(inst)


def test_nested_binary_config():
    input_data1 = {"name": "Alex", "age": 42}
    input_data2 = {"name": "Sam", "age": 24}
    data1 = struct.pack("32s", json.dumps(input_data1).encode())
    data2 = struct.pack("32s", yaml.dump(input_data2).encode())
    data3 = struct.pack("32s", tomli_w.dumps(input_data2).encode())

    inst = NestedBinaryConfig.load({"data2": data2, "data1": data1, "data3": data3})

    assert inst.data1.name != inst.data2.name

    assert inst._pack() == data1 + data2 + data3 == asbytes(inst)


class Version(BinaryConfig):
    major = BinaryField(int)
    minor = BinaryField(int)
    patch = BinaryField(int)


class Versions(BinaryConfig):
    first = BinaryField(Version)
    second = BinaryField(Version)


def test_binary_config_with_external_block():
    v1 = struct.pack("i i i", 1, 12, 5)
    v2 = struct.pack("i i i", 0, 4, 2)
    data = Versions.load(v1 + v2)

    assert data.first.patch == 5
    assert data.second.major == 0

    assert data._pack() == v1 + v2 == asbytes(data)


class IsNumber(BinaryConfig):
    value = BinaryField(int, format="h")


class IsBigNumber(BinaryConfig):
    value = BinaryField(int, format="l")


class HasNumber(BinaryConfig):
    contains = BinaryField(IsNumber)


def test_resize_config():
    inst = HasNumber()
    assert inst._get_length() == 2
    inst.contains = 12
    assert inst._get_length() == 2

    big_num = IsBigNumber()
    big_num.value = 420

    inst.contains = big_num
    assert inst._get_length() == 8



from pathlib import Path
from typing import Optional

from src.configuraptor import load_into, postpone


class PathConfig:
    required: Path
    optional: Optional[Path]


class PathConfigPostponed:
    required: Path = postpone()
    optional: Optional[Path] = postpone()


def test_regular():
    conf = load_into(
        PathConfig,
        {
            "required": Path.home(),
        },
    )

    assert conf.required == Path.home()
    assert isinstance(conf.required, Path)


def test_postponed():
    conf = load_into(PathConfigPostponed, {})

    assert conf

    conf.required = Path.home()

    assert conf.required == Path.home()



import datetime as dt
import math
import typing

import pytest

from src import configuraptor
from src.configuraptor.errors import (
    ConfigError,
    ConfigErrorInvalidType,
    ConfigErrorMissingKey,
)
from src.configuraptor.helpers import find_pyproject_toml

from .constants import EMPTY_FILE, EXAMPLE_FILE, PYTEST_EXAMPLES, _load_toml


def test_example_is_valid_toml():
    data = _load_toml()

    assert data


class AbsHasName:
    # can be inherited by other classes with a 'name' attribute.
    name: str


# class FirstExtraName:
#     first: str
#     last: str
#
#
# class FirstExtraPoint:
#     x: int
#     y: int
#
#
# class FirstExtraAnimalType(AbsHasName):
#     ...
#
#
# class FirstExtraAnimal:
#     type: FirstExtraAnimalType
#
#
# class FirstExtra:
#     name: FirstExtraName
#     point: FirstExtraPoint
#     animal: FirstExtraAnimal


class First:
    string: str
    list_of_string: list[str]
    list_of_int: list[int]
    list_of_float: list[float]
    list_of_numbers: list[float | int]
    some_boolean: bool
    number: float | int
    not_a_number: math.nan
    datetime: dt.datetime
    datetimes: list[dt.datetime]
    extra: typing.Optional[dict[str, typing.Any]]


class FruitDetails:
    color: str
    shape: str


class FruitVariety(AbsHasName): ...


class Fruit(AbsHasName):
    varieties: list[FruitVariety]
    physical: typing.Optional[FruitDetails]


class SecondExtra:
    allowed: bool


class Tool:
    first: First
    fruits: list[Fruit]
    second_extra: SecondExtra


class Empty:
    default: str = "allowed"


def test_empty():
    empty = configuraptor.load_into(Empty, {})
    assert empty and empty.default == "allowed"

    with pytest.raises(ConfigError):
        configuraptor.load_into(Tool, {})

    with pytest.raises(ConfigError):
        configuraptor.load_into(First, EMPTY_FILE, key="tool.first")


class MyConfig:
    public_key: str
    private_key: str
    extra: int


def test_multiple_files():
    public_file = str(PYTEST_EXAMPLES / "my_config.toml")
    private_file = str(PYTEST_EXAMPLES / "my_secrets.env")

    config = configuraptor.load_into(
        MyConfig,
        [public_file, private_file, {"extra": 3}],  # toml  # .env  # raw dict
        key="my_config.custom",  # should work even if only relevant for toml file
        # lower keys is automatically set to True
    )

    assert config.public_key == "this is public"
    assert config.private_key == "THIS IS PRIVATE" != "<overwite me>"


def test_basic_classes():
    data = _load_toml()

    tool = configuraptor.load_into(Tool, data)
    first = configuraptor.load_into(First, EXAMPLE_FILE, key="tool.first")

    assert tool.first.extra["name"]["first"] == first.extra["name"]["first"]


class Relevant:
    key: str


class OptionalRelevant:
    key: typing.Optional[str]


class Irrelevant:
    key: dict


def test_guess_key_from_multiple_keys():
    file = str(PYTEST_EXAMPLES / "with_multiple_toplevel_keys.toml")
    relevant = configuraptor.load_into(Relevant, file)
    assert relevant and relevant.key == "this one!"


def test_guess_key_no_match():
    file = str(PYTEST_EXAMPLES / "with_multiple_toplevel_keys.toml")
    # no key and no match by class name:
    relevant = configuraptor.load_into(Irrelevant, file)
    assert relevant and relevant.key["value"] == "fallback"


class TooLong:
    type: str


def test_invalid_type():
    file = str(PYTEST_EXAMPLES / "with_multiple_toplevel_keys.toml")
    with pytest.raises(ConfigErrorInvalidType):
        # tool.key contains an int instead of a str
        configuraptor.load_into(Relevant, file, key="tool")

    with pytest.raises(ConfigErrorInvalidType):
        # complex.type is a long dict but is typed as string.
        # it will be truncated in the error message
        try:
            configuraptor.load_into(TooLong, file, key="complex")
        except ConfigErrorInvalidType as e:
            assert "..." in str(e)
            raise e


def test_missing_key():
    file = str(PYTEST_EXAMPLES / "with_multiple_toplevel_keys.toml")
    with pytest.raises(ConfigErrorMissingKey):
        # key.key does not exist
        try:
            configuraptor.load_into(Relevant, file, key="key")
        except ConfigErrorMissingKey as e:
            assert "Config key 'key'" in str(e)
            raise e

    # allowed here since 'key' is optional:
    inst = configuraptor.load_into(OptionalRelevant, file, key="key")
    assert inst.key is None


class Point:
    x: int
    y: int


class Structure:
    # contents: dict[str, Point]
    contents: dict[str, Point]


def test_dict_of_custom():
    file = PYTEST_EXAMPLES / "with_dict_of_custom.toml"

    structure = configuraptor.load_into(Structure, file)
    for key, value in structure.contents.items():
        assert isinstance(key, str)
        assert isinstance(value, Point)


try:
    import contextlib

    chdir = contextlib.chdir
except AttributeError:
    from contextlib_chdir import chdir


class ProjectToml:
    my_key: str


def test_pyproject_toml():
    with chdir("pytest_examples/nested"):
        assert find_pyproject_toml()

        config = configuraptor.load_into(ProjectToml, key="tool.configuraptor.test")

    assert config.my_key == "my_value"

    with chdir("/"):
        assert not find_pyproject_toml()


class ShouldHaveListOfString:
    required: list[str]
    not_required: typing.Optional[list[str]]


def test_missing_required_parameterized():
    # should go through is_optional -> is_parameterized -> typing.get_origin(_type) in (dict, list) -> False -> Error
    data = {"not_required": ["list", "of", "string"]}
    with pytest.raises(ConfigErrorMissingKey):
        configuraptor.load_into(ShouldHaveListOfString, data)



import json
import typing

import pytest
import tomli_w
import yaml

from src.configuraptor import TypedConfig
from src.configuraptor.dump import asdict, asjson, astoml, asyaml


class Simple(TypedConfig):
    key: str
    num: int


data = {
    "simple": {
        "key": "abc",
        "num": 123,
    }
}

simple = Simple.load(data)


def test_asdict_simple():
    assert asdict(simple) == data


def test_asjson():
    assert asjson(simple) == json.dumps(data)


def test_asyaml():
    assert asyaml(simple) == yaml.dump(data)


def test_astoml():
    assert astoml(simple) == tomli_w.dumps(data)


class Dependency:
    name: str


class Complex(TypedConfig):
    name: str
    dependency: Dependency
    dependencies: list[Dependency]
    extra: dict[str, Dependency]

    def update_internals(self):
        self._internal = "weird"
        self.__callable = self.update_internals


data_complex = {
    "complex": {
        "name": "hi",
        "dependency": {"name": "hey"},
        "dependencies": [{"name": "he"}, {"name": "ha"}],
        "extra": {"first": {"name": "1"}, "second": {"name": "2"}},
        "ignored": "should not exist in asdict",
    }
}

complex = Complex.load(data_complex)


def test_asdict_complex():
    complex.update_internals()

    with pytest.raises(Exception):
        # because cls.__callable is not json serializable
        asjson(complex, exclude_internals=0)

    data_complex["complex"].pop("ignored")
    json_complex = json.dumps(data_complex)
    assert asjson(complex, exclude_internals=2) == json_complex

    data_complex["complex"]["_internal"] = "weird"
    json_complex = json.dumps(data_complex)

    assert asjson(complex, exclude_internals=1) == json_complex


# def test_asdict_dataclass():
#     ...



"""
Example with dataclasses
"""

from dataclasses import dataclass, field

from configuraptor import load_into


@dataclass
class Two:
    name: str
    some_str: str
    some_int: int
    include: list[str] = field(default_factory=list)  # <- not required in config file since it has a default value


@dataclass
class Simple:
    name: str
    two: Two


def main() -> None:
    data = {"simple": {"name": "Steve", "two": {"name": "Alex", "some_str": "string", "some_int": 30}}}

    simple = load_into(Simple, data)
    two = simple.two
    print(simple, two)


if __name__ == "__main__":
    main()



from configuraptor import load_into_instance


class Person:
    name: str
    age: int

    def __init__(self, name: str, age: int = None):
        self.name = name
        if age is not None:
            self.age = age
        # else: age is still unknown


maria = Person("Maria", 45)
steve = Person("Steve")

load_into_instance(maria, "ages.toml", key="maria")
load_into_instance(steve, "ages.toml", key="steve")

print(maria.age, steve.age)
# 45, 24



"""
Contains module-specific custom errors.
"""

import typing
from dataclasses import dataclass


class ConfigError(Exception):
    """
    Base exception class for this module.
    """


# class ConfigErrorGroup(ConfigError, ExceptionGroup):
#     """
#     Base Exception class for this module, but for exception groups (3.11+)
#     """
#     def __init__(self, _type: str, errors: list[Exception]):
#         more = len(errors) > 1
#         cnt = "Multiple" if more else "One"
#         s = "s" if more else ""
#         message = f"{cnt} {_type}{s} in config!"
#         super().__init__(message, errors)
#         if not errors:
#             raise ValueError("Error group raised without any errors?")


@dataclass
class ConfigErrorMissingKey(ConfigError):
    """
    Exception for when the config file is missing a required key.
    """

    key: str
    cls: type
    annotated_type: type

    def __post_init__(self) -> None:
        """
        Automatically filles in the names of annotated type and cls for printing from __str__.
        """
        self._annotated_type = getattr(self.annotated_type, "__name__", str(self.annotated_type))
        self._cls = self.cls.__name__

    def __str__(self) -> str:
        """
        Custom error message based on dataclass values and calculated actual type.
        """
        return (
            f"Config key '{self.key}' (type `{self._annotated_type}`) "
            f"of class `{self._cls}` was not found in the config, "
            f"but is required as a default value is not specified."
        )


@dataclass
class ConfigErrorExtraKey(ConfigError):
    """
    Exception for when the config file is missing a required key.
    """

    key: str
    value: str
    cls: type

    def __post_init__(self) -> None:
        """
        Automatically filles in the names of annotated type and cls for printing from __str__.
        """
        self._cls = self.cls.__name__
        self._type = type(self.value)

    def __str__(self) -> str:
        """
        Custom error message based on dataclass values and calculated actual type.
        """
        return (
            f"Config key '{self.key}' (value: `{self.value}` type `{self._type}`) "
            f"does not exist on class `{self._cls}`, but was attempted to be updated. "
            f"Use strict = False to allow this behavior."
        )


@dataclass
class ConfigErrorCouldNotConvert(ConfigError):
    """
    Raised by `convert_between` if something funky is going on (incompatible types etc.).
    """

    from_t: type
    to_t: type
    value: typing.Any

    def __str__(self) -> str:
        """
        Custom error message based on dataclass values and calculated actual type.
        """
        return f"Could not convert `{self.value}` from `{self.from_t}` to `{self.to_t}`"


@dataclass
class ConfigErrorInvalidType(ConfigError):
    """
    Exception for when the config file contains a key with an unexpected type.
    """

    key: str
    value: typing.Any
    expected_type: type

    def __post_init__(self) -> None:
        """
        Store the actual type of the config variable.
        """
        self.actual_type = type(self.value)

        max_len = 50
        self._value = str(self.value)
        if len(self._value) > max_len:
            self._value = f"{self._value[:max_len]}..."

    def __str__(self) -> str:
        """
        Custom error message based on dataclass values and calculated actual type.
        """
        return (
            f"Config key '{self.key}' had a value (`{self._value}`) with a type (`{self.actual_type}`) "
            f"that was not expected: `{self.expected_type}` is the required type."
        )


@dataclass
class ConfigErrorImmutable(ConfigError):
    """
    Raised when an immutable Mapping is attempted to be updated.
    """

    cls: type

    def __post_init__(self) -> None:
        """
        Store the class name.
        """
        self._cls = self.cls.__name__

    def __str__(self) -> str:
        """
        Custom error message.
        """
        return f"{self._cls} is Immutable!"


@dataclass
class IsPostponedError(ConfigError):
    """
    Error thrown when you try to access a 'postponed' property without filling its value first.
    """

    message: str = "This postponed property has not been filled yet!"



from typing import Optional

import pytest

from src.configuraptor import TypedConfig, alias, load_into, postpone
from src.configuraptor.alias import is_alias
from src.configuraptor.core import check_and_convert_type
from src.configuraptor.errors import ConfigError


class MyConfig:
    key_one: str
    key_two: str = alias("key_one")


class MyConfigInvalid(TypedConfig):
    key_three: str = alias("key_four")
    key_four: str = alias("key_three")


class AliasWithPostponed(TypedConfig):
    first: str = postpone()
    second: str = alias("first")


class MyConfigInvalidTwo(TypedConfig):
    key_five: str = alias("key_five")


class UnresolvedOptionalAlias(TypedConfig):
    fro: Optional[str]
    to: Optional[str] = alias("fro")


def test_it():
    conf1 = load_into(MyConfig, {"key_one": "one"})

    assert conf1.key_one == conf1.key_two == "one"

    conf2 = load_into(MyConfig, {"key_two": "two"})

    assert conf2.key_one == conf2.key_two == "two"

    # normal one should have priority over alias field:
    conf3 = load_into(MyConfig, {"key_one": "ONE", "key_two": "two"})

    assert conf3.key_one == "ONE"

    conf = UnresolvedOptionalAlias.load()

    assert conf.fro is None
    assert conf.to is None

    assert check_and_convert_type(alias("something"), Optional[str], False) is None


def test_with_postpone():
    c = AliasWithPostponed.load({"first": "one"})

    assert c.first == c.second == "one"

    c.update(first="two")

    assert c.first == c.second == "two"

    c.update(second="two2")

    assert c.first == c.second == "two2"

    c = AliasWithPostponed.load({})
    c.first = "three"
    assert c.first == c.second == "three"

    c.second = "three2"
    assert c.first == c.second == "three2"


def test_invalid():
    with pytest.raises(ConfigError):
        load_into(MyConfigInvalid, {})

    assert load_into(MyConfigInvalid, {"key_three": "3"})
    assert load_into(MyConfigInvalid, {"key_four": "4"})

    with pytest.raises(ConfigError):
        load_into(MyConfigInvalidTwo, {})

    assert load_into(MyConfigInvalidTwo, {"key_five": "123"})


def test_is_alias():
    assert not is_alias(UnresolvedOptionalAlias, "fro")
    assert is_alias(UnresolvedOptionalAlias, "to")



"""
Contains stand-alone helper functions.
"""

import contextlib
import dataclasses as dc
import io
import math
import types
import typing
from collections import ChainMap
from pathlib import Path

from typeguard import TypeCheckError
from typeguard import check_type as _check_type

try:
    import annotationlib
except ImportError:  # pragma: no cover
    annotationlib = None


def camel_to_snake(s: str) -> str:
    """
    Convert CamelCase to snake_case.

    Source:
        https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    """
    return "".join([f"_{c.lower()}" if c.isupper() else c for c in s]).lstrip("_")


# def find_pyproject_toml() -> typing.Optional[str]:
#     """
#     Find the project's config toml, looks up until it finds the project root (black's logic).
#     """
#     return black.files.find_pyproject_toml((os.getcwd(),))


def find_pyproject_toml(start_dir: typing.Optional[Path | str] = None) -> Path | None:
    """
    Search for pyproject.toml starting from the current working directory \
     and moving upwards in the directory tree.

    Args:
        start_dir: Starting directory to begin the search.
            If not provided, uses the current working directory.

    Returns:
        Path or None: Path object to the found pyproject.toml file, or None if not found.
    """
    start_dir = Path.cwd() if start_dir is None else Path(start_dir).resolve()

    current_dir = start_dir

    while str(current_dir) != str(current_dir.root):
        pyproject_toml = current_dir / "pyproject.toml"
        if pyproject_toml.is_file():
            return pyproject_toml
        current_dir = current_dir.parent

    # If not found anywhere
    return None


Type = typing.Type[typing.Any]


def _cls_annotations(c: type) -> dict[str, type]:  # pragma: no cover
    """
    Functions to get the annotations of a class (excl inherited, use _all_annotations for that).

    Uses `annotationlib` if available (since 3.14) and if so, resolves forward references immediately.
    """
    if annotationlib:
        return typing.cast(
            dict[str, type], annotationlib.get_annotations(c, format=annotationlib.Format.VALUE, eval_str=True)
        )
    else:
        # note: idk why but this is not equivalent (the first doesn't work well):
        # return getattr(c, "__annotations__", {})
        return c.__dict__.get("__annotations__") or {}


def _all_annotations(cls: type) -> ChainMap[str, type]:
    """
    Returns a dictionary-like ChainMap that includes annotations for all \
    attributes defined in cls or inherited from superclasses.
    """
    # chainmap reverses the iterable, so reverse again beforehand to keep order normally:

    return ChainMap(*(_cls_annotations(c) for c in getattr(cls, "__mro__", [])))


def all_annotations(cls: Type, _except: typing.Iterable[str] = None) -> dict[str, type[object]]:
    """
    Wrapper around `_all_annotations` that filters away any keys in _except.

    It also flattens the ChainMap to a regular dict.
    """
    if _except is None:
        _except = set()

    _all = _all_annotations(cls)
    return {k: v for k, v in _all.items() if k not in _except}


T = typing.TypeVar("T")


def check_type(value: typing.Any, expected_type: typing.Type[T]) -> typing.TypeGuard[T]:
    """
    Given a variable, check if it matches 'expected_type' (which can be a Union, parameterized generic etc.).

    Based on typeguard but this returns a boolean instead of returning the value or throwing a TypeCheckError
    """
    try:
        _check_type(value, expected_type)
        return True
    except TypeCheckError:
        return False


def is_builtin_type(_type: Type) -> bool:
    """
    Returns whether _type is one of the builtin types.
    """
    return _type.__module__ in ("__builtin__", "builtins")


# def is_builtin_class_instance(obj: typing.Any) -> bool:
#     return is_builtin_type(obj.__class__)


def is_from_types_or_typing(_type: Type) -> bool:
    """
    Returns whether _type is one of the stlib typing/types types.

    e.g. types.UnionType or typing.Union
    """
    return _type.__module__ in ("types", "typing")


def is_from_other_toml_supported_module(_type: Type) -> bool:
    """
    Besides builtins, toml also supports 'datetime' and 'math' types, \
    so this returns whether _type is a type from these stdlib modules.
    """
    return _type.__module__ in ("datetime", "math")


def is_parameterized(_type: Type) -> bool:
    """
    Returns whether _type is a parameterized type.

    Examples:
        list[str] -> True
        str -> False
    """
    return typing.get_origin(_type) is not None


def is_custom_class(_type: Type) -> bool:
    """
    Tries to guess if _type is a builtin or a custom (user-defined) class.

    Other logic in this module depends on knowing that.
    """
    return (
        type(_type) is type
        and not is_builtin_type(_type)
        and not is_from_other_toml_supported_module(_type)
        and not is_from_types_or_typing(_type)
    )


def instance_of_custom_class(var: typing.Any) -> bool:
    """
    Calls `is_custom_class` on an instance of a (possibly custom) class.
    """
    return is_custom_class(var.__class__)


def is_optional(_type: Type | typing.Any) -> bool:
    """
    Tries to guess if _type could be optional.

    Examples:
        None -> True
        NoneType -> True
        typing.Union[str, None] -> True
        str | None -> True
        list[str | None] -> False
        list[str] -> False
    """
    if _type and (is_parameterized(_type) and typing.get_origin(_type) in (dict, list)) or (_type is math.nan):
        # e.g. list[str]
        # will crash issubclass to test it first here
        return False

    try:
        return (
            _type is None
            or types.NoneType in typing.get_args(_type)  # union with Nonetype
            or issubclass(types.NoneType, _type)
            or issubclass(types.NoneType, type(_type))  # no type  # Nonetype
        )
    except TypeError:
        # probably some weird input that's not a type
        return False


def dataclass_field(cls: Type, key: str) -> typing.Optional[dc.Field[typing.Any]]:
    """
    Get Field info for a dataclass cls.
    """
    fields = getattr(cls, "__dataclass_fields__", {})
    return fields.get(key)


@contextlib.contextmanager
def uncloseable(fd: typing.BinaryIO) -> typing.Generator[typing.BinaryIO, typing.Any, None]:
    """
    Context manager which turns the fd's close operation to no-op for the duration of the context.
    """
    close = fd.close
    fd.close = lambda: None  # type: ignore
    yield fd
    fd.close = close  # type: ignore


def as_binaryio(file: str | Path | typing.BinaryIO | None, mode: typing.Literal["rb", "wb"] = "rb") -> typing.BinaryIO:
    """
    Convert a number of possible 'file' descriptions into a single BinaryIO interface.
    """
    if isinstance(file, str):
        file = Path(file)
    if isinstance(file, Path):
        file = file.open(mode)
    if file is None:
        file = io.BytesIO()
    if isinstance(file, io.BytesIO):
        # so .read() works after .write():
        file.seek(0)
        # so the with-statement doesn't close the in-memory file:
        file = uncloseable(file)  # type: ignore

    return file



"""
Contains the Abstract config class shared by TypedConfig and BinaryConfig.
"""

import os
import types
import typing
from pathlib import Path

import dotenv
from dotenv import find_dotenv

# T is a reusable typevar
T = typing.TypeVar("T")
# t_typelike is anything that can be type hinted
T_typelike: typing.TypeAlias = type | types.UnionType  # | typing.Union
# t_data is anything that can be fed to _load_data
T_data_types = str | Path | bytes | dict[str, typing.Any] | None
T_data = T_data_types | list[T_data_types]

# c = a config class instance, can be any (user-defined) class
C = typing.TypeVar("C")
# type c is a config class
Type_C = typing.Type[C]

AnyType: typing.TypeAlias = typing.Type[typing.Any]
T_Type = typing.TypeVar("T_Type", bound=AnyType)


class AbstractTypedConfig:
    """
    These functions only exist on the class, not on instances.
    """

    @classmethod
    def load(
        cls: typing.Type[C],
        data: T_data = None,
        key: str = None,
        init: dict[str, typing.Any] = None,
        strict: bool = True,
        lower_keys: bool = False,
        convert_types: bool = False,
    ) -> C:
        """
        Load a class' config values from the config file.

        SomeClass.load(data, ...) = load_into(SomeClass, data, ...).
        """
        from .core import load_into

        return load_into(
            cls, data, key=key, init=init, strict=strict, lower_keys=lower_keys, convert_types=convert_types
        )

    @classmethod
    def from_env(
        cls: typing.Type[C],
        load_dotenv: str | bool = False,
        init: dict[str, typing.Any] = None,
        strict: bool = True,
        convert_types: bool = True,
    ) -> C:
        """
        Create an instance of the typed config class by loading environment variables and initializing \
            object attributes based on those values.

        Args:
            cls (typing.Type[C]): The class to create an instance of.
            init (dict[str, typing.Any], optional): Additional initialization data to be used
                in the object creation. Defaults to None.
            strict (bool, optional): If True, raise an error if any required environment variable
                is missing. Defaults to True.
            convert_types (bool, optional): If True, attempt to convert environment variable values
                to the appropriate Python types. Defaults to False.
            load_dotenv (str | bool, optional): Path to a dotenv file or True to load the default
                dotenv file. If False, no dotenv file will be loaded. Defaults to False.

        Returns:
            C: An instance of the class `C` with attributes initialized based on the environment variables.
        """
        from .core import load_into

        if load_dotenv:
            dotenv_path = load_dotenv if isinstance(load_dotenv, str) else find_dotenv(usecwd=True)
            dotenv.load_dotenv(dotenv_path)

        data = {**os.environ}

        return load_into(cls, data, lower_keys=True, init=init, strict=strict, convert_types=convert_types)



import typing

import pytest

from src.configuraptor import asdict, asjson, load_into


class Test: ...


@pytest.mark.mypy_testing
def mypy_test_invalid_assignment() -> None:
    test = load_into(Test, {})

    typing.reveal_type(test)  # R: tests.test_mypy.Test


@pytest.mark.mypy_testing
def mypy_test_asjson() -> None:
    jsonified = asjson("-")

    typing.reveal_type(jsonified)  # R: builtins.str


@pytest.mark.mypy_testing
def mypy_test_asdict() -> None:
    dictified = asdict(Test)

    typing.reveal_type(dictified)  # R: builtins.dict[builtins.str, Any]



from src.configuraptor import TypedConfig, beautify


@beautify
class MyClass(TypedConfig):
    required: str
    optional: str | None


config = MyClass.load({"required": "there"})


def test_required():
    assert config.required == "there"
    config.required = None
    assert config.required == "there"


def test_optional():
    assert config.optional is None
    config.optional = "not none"
    assert config.optional == "not none"
    config.optional = None  # should be allowed according to the annotation
    assert config.optional is None



"""
Exposes `register_loader` to define loader for specific file types.
"""

import typing
from pathlib import Path

from ._types import T_config

T_loader = typing.Callable[[typing.BinaryIO, Path], T_config]
T_WrappedLoader = typing.Callable[[T_loader], T_loader]

T_dumper = typing.Callable[..., str | dict[str, typing.Any]]
T_WrappedDumper = typing.Callable[[T_dumper], T_dumper]

LOADERS: dict[str, T_loader] = {}
DUMPERS: dict[str, T_dumper] = {}

R = typing.TypeVar("R")
AnyCallable = typing.Callable[..., typing.Any]


def register_something(storage: dict[str, typing.Any], *extension_args: typing.Any) -> AnyCallable:
    f_outer = None
    extension_set = set()

    for extension in extension_args:
        if not isinstance(extension, str):
            f_outer = extension
            extension = extension.__name__

        elif extension.startswith("."):
            extension = extension.removeprefix(".")

        extension_set.add(extension)

    def wrapper(f_inner: typing.Callable[..., R]) -> typing.Callable[..., R]:
        storage.update({ext: f_inner for ext in extension_set})
        return f_inner

    if f_outer:
        return wrapper(f_outer)  # -> T_Loader
    else:
        return wrapper  # -> T_WrappedLoader


@typing.overload
def register_loader(*extension_args: str) -> T_WrappedLoader:
    """
    Overload for case with parens.

    @register_loader("yaml", ".yml")
    def load_yaml(...):
        ...

    # extension_args is a tuple of strings
    # this will return a wrapper which takes `load_yaml` as input and output.
    """


@typing.overload
def register_loader(*extension_args: T_loader) -> T_loader:
    """
    Overload for case without parens.

    @register_loader
    def json(...):
        ...

    # extension_args is a tuple of 1: `def json`
    # this will simply return the `json` method itself.
    """


def register_loader(*extension_args: str | T_loader) -> T_loader | T_WrappedLoader:
    """
    Register a data loader for a new filetype.

    Used as a decorator on a method that takes two arguments:
    (BinaryIO, Path) - an open binary file stream to the config file and the pathlib.Path to the config file.
    By default, the open file handler can be used.
    However, some loaders (such as .ini) don't support binary file streams.
    These can use the Path to open and read the file themselves however they please.
    """
    return register_something(LOADERS, *extension_args)


@typing.overload
def register_dumper(*extension_args: str) -> typing.Callable[[typing.Callable[..., R]], typing.Callable[..., R]]:
    """
    Overload for case with parens.

    @register_dumper("yaml", ".yml")
    def dump_yaml(...):
        ...

    # extension_args is a tuple of strings
    # this will return a wrapper which takes `dump_yaml` as input and output.
    """


@typing.overload
def register_dumper(*extension_args: typing.Callable[..., R]) -> typing.Callable[..., R]:
    """
    Overload for case without parens.

    @register_dumper
    def json(...):
        ...

    # extension_args is a tuple of 1: `def json`
    # this will simply return the `json` method itself.
    """


def register_dumper(
    *extension_args: str | typing.Callable[..., R],
) -> typing.Callable[[typing.Callable[..., R]], typing.Callable[..., R]] | typing.Callable[..., R]:
    """
    Register a data dumper for a new filetype.

    Not everything that can be loaded, can also be dumped (currently).
    """
    return register_something(DUMPERS, *extension_args)


__all__ = ["register_loader", "register_dumper", "LOADERS", "DUMPERS"]



"""
Logic to do with parsing bytestrings as configuration (using struct).
"""

import collections
import struct
import typing
from dataclasses import dataclass
from io import BytesIO
from pathlib import Path

from typing_extensions import Self

from . import loaders
from .abs import AbstractTypedConfig
from .helpers import is_custom_class
from .loaders.register import DUMPERS

BINARY_TYPES = typing.Union[str, float, int, bool]


class BinaryConfig(AbstractTypedConfig):
    """
    Inherit this class if you want your config or a section of it to be parsed using struct.
    """

    _fields: collections.OrderedDict[str, "_BinaryField"]

    def __init__(self) -> None:
        """
        Before filling the class with data, we store the fields (BinaryField) for later use.
        """
        fields, elements = self._collect_fields()
        self._fields = collections.OrderedDict(zip(fields, elements))
        super().__init__()

    @classmethod
    def _collect_fields(cls) -> tuple[list[str], list["_BinaryField"]]:
        """
        Get the class' field names and dataclass instances.
        """
        elements: list[_BinaryField] = []
        fields: list[str] = []

        for field, value in cls.__dict__.items():
            if field.startswith("_"):
                continue
            if not isinstance(value, _BinaryField):
                # other data, skip
                continue

            fields.append(field)
            elements.append(value)

        return fields, elements

    @classmethod
    def _parse(cls, data: bytes | dict[str, bytes]) -> dict[str, BINARY_TYPES]:
        """
        Parse a bytestring or a dict of bytestrings (in the right order).
        """
        from .core import load_into

        # NOTE: annotations not used!
        fields, elements = cls._collect_fields()

        if isinstance(data, dict):
            # create one long bytestring of data in the right order:
            data = b"".join(data[field] for field in fields)

        unpacked = struct.unpack(" ".join(str(_) for _ in elements), data)
        final_data: dict[str, BINARY_TYPES] = {}

        zipped: typing.Iterable[tuple[str, typing.Any, _BinaryField]] = zip(fields, unpacked, elements)
        for field, value, meta in zipped:
            if isinstance(value, bytes) and not issubclass(meta.klass, BinaryConfig):
                value = value.strip(b"\x00").decode()

            if meta.special:
                # e.g. load from JSON
                value = meta.special(value)

            # ensure it's the right class (e.g. bool):
            value = load_into(meta.klass, value) if is_custom_class(meta.klass) else meta.klass(value)

            final_data[field] = value

        return final_data

    @classmethod
    def _parse_into(cls, data: bytes | dict[str, bytes]) -> Self:
        """
        Create a new instance based on data.
        """
        converted = cls._parse(data)
        inst = cls()
        inst.__dict__.update(**converted)
        return inst

    def _pack(self) -> bytes:
        """
        Pack an instance back into a bytestring.
        """
        fmt = " ".join(str(_) for _ in self._fields.values())

        values = [self._fields[k].pack(v) for k, v in self.__dict__.items() if not k.startswith("_")]

        return struct.pack(fmt, *values)

    @classmethod
    def _format(cls) -> str:
        _, fields = cls._collect_fields()

        return " ".join(str(_) for _ in fields)

    @classmethod
    def _get_length(cls) -> int:
        """
        How many bytes do the fields of this class have?
        """
        fmt = cls._format()

        return struct.calcsize(fmt)

    def __setattr__(self, key: str, value: typing.Any) -> None:
        """
        When setting a new field for this config, update the _fields property to have the correct new type + size.
        """
        if not key.startswith("_") and isinstance(value, BinaryConfig):
            field = self._fields[key]
            field.klass = value.__class__
            field.length = value.__class__._get_length()

        return super().__setattr__(key, value)


@dataclass(slots=True)
class _BinaryField:
    """
    Class that stores info to parse the value from a bytestring.

    Returned by BinaryField, but overwritten on instances with the actual value of type klass.
    """

    klass: typing.Type[typing.Any]
    length: int
    fmt: str
    special: typing.Callable[[typing.Any], dict[str, typing.Any]] | None
    packer: typing.Callable[[typing.Any], typing.Any] | None

    def __str__(self) -> str:
        return f"{self.length}{self.fmt}"

    def pack(self, value: typing.Any) -> typing.Any:
        if self.packer:
            value = self.packer(value)
        if isinstance(value, str):
            return value.encode()
        if isinstance(value, BinaryConfig):
            return value._pack()
        return value


T = typing.TypeVar("T")

# https://docs.python.org/3/library/struct.html
# DEFAULT_LENGTHS = {
#     "x": 1,
#     "c": 1,
#     "b": 1,
#     "?": 1,
#     "h": 2,
#     "H": 2,
#     "i": 4,
#     "I": 4,
#     "l": 4,
#     "L": 4,
#     "q": 8,
#     "Q": 8,
#     "n": 8,
#     "N": 8,
#     "e": 2,
#     "f": 4,
#     "d": 8,
#     "s": 1,
#     "p": 1,
#     "P": 8,
# }

DEFAULT_FORMATS = {
    str: "s",
    int: "i",
    float: "f",
    bool: "?",  # b
}


def BinaryField(klass: typing.Type[T], **kw: typing.Any) -> T:
    """
    Fields for BinaryConfig can not be annotated like a regular typed config, \
    because more info is required (such as struct format/type and length).

    This actually returns a _BinaryField but when using load/load_into, the value will be replaced with type 'klass'.

    Args:
        klass (type): the final type the value will have
        format (str): either one of the formats of struct (e.g. 10s) or a loadable format (json, toml, yaml etc.)
        length (int): how many bytes of data to store? (required for str, unless you enter it in format already)

    Usage:
        class MyConfig(BinaryConfig):
            string = BinaryField(str, length=5) # string of 5 characters
            integer = BinaryField(int)
            complex = BinaryField(OtherClass, format='json', length=64)
                        # will extract 64 bytes of string and try to convert to the linked class
                        # (using regular typeconfig logic)
    """
    special = None
    packer = None

    if issubclass(klass, BinaryConfig):
        fmt = "s"  # temporarily group as one string
        length = kw.get("length", klass._get_length())
    else:
        fmt = kw.get("format") or DEFAULT_FORMATS[klass]
        if loader := loaders.get(fmt, None):
            special = lambda data: loader(  # noqa: E731
                BytesIO(data if isinstance(data, bytes) else data.encode()), Path()
            )
            if _packer := DUMPERS.get(fmt, None):
                packer = lambda data: _packer(data, with_top_level_key=False)  # noqa: E731
            length = kw["length"]
            fmt = "s"
        elif len(fmt) > 1:
            # length in format: 10s
            length, fmt = int(fmt[:-1]), fmt[-1]
        else:
            length = kw.get("length", 1)

    field = _BinaryField(
        klass,
        fmt=fmt,
        length=length,
        special=special,
        packer=packer,
    )

    return typing.cast(T, field)



import json
import typing

import pytest

from src.configuraptor import TypedConfig, load_into
from src.configuraptor.beautify import beautify


@beautify
class MyClass:
    string: str
    number: int
    _internal: bool = False


@beautify(repr=True, str=False)
class MyOtherClass:
    number: float
    boolean: bool


@beautify()
class CustomRepr:
    some: str

    def __repr__(self) -> str:
        return "dont-touch-me"

    def __str__(self) -> str:
        return "dont-touch-me-either"


class FromTypedConfig(TypedConfig):
    # should be auto-beautified
    another: bool


class NoBeautify(TypedConfig, beautify=False):
    # should NOT be auto-beautified
    beautiful: bool


test1 = load_into(MyClass, {"string": "123", "number": 123})
test2 = load_into(MyOtherClass, {"number": 12.3, "boolean": True})
test3 = load_into(CustomRepr, {"some": "value"})


@pytest.mark.mypy_testing
def mypy_test_decorated_cls() -> None:
    typing.reveal_type(test1)  # R: tests.test_beautify.MyClass
    typing.reveal_type(test2)  # R: tests.test_beautify.MyOtherClass


def test_str() -> None:
    assert json.loads(str(test1))

    with pytest.raises(json.JSONDecodeError):
        # str=False so it shouldn't be json-able
        assert json.loads(str(test2))

    assert str(test3) == "dont-touch-me-either"


def test_repr() -> None:
    assert "MyClass" in repr(test1)
    assert "string" in repr(test1)
    assert "123" in repr(test1)
    assert "_internal" not in repr(test1)
    assert "_internal" not in str(test1)

    assert "MyOtherClass" in repr(test2)
    assert "number" in repr(test2)
    assert "True" in repr(test2)

    assert repr(test3) == "dont-touch-me"


def test_from_typedconfig() -> None:
    inst = FromTypedConfig.load({"another": True})

    assert "<FromTypedConfig" in repr(inst)
    assert "another" in repr(inst)
    assert "True" in repr(inst)

    assert "<FromTypedConfig" not in str(inst)
    assert "another" in str(inst)
    assert "true" in str(inst)
    assert json.loads(str(inst))


def test_nobeautify() -> None:
    ugly = NoBeautify.load({"beautiful": False})

    assert "<NoBeautify" not in repr(ugly)
    assert "beautiful" not in repr(ugly)
    assert "False" not in repr(ugly)

    assert "<NoBeautify" not in str(ugly)
    assert "beautiful" not in str(ugly)
    assert "false" not in str(ugly)

    with pytest.raises(json.JSONDecodeError):
        assert not json.loads(str(ugly))



from configuraptor import Singleton, load_into


class MyConfig(Singleton):
    string: str
    number: int

    def update(self, string: str, number: int):
        self.string = string
        self.number = number

    def __repr__(self):
        return f"{self.string=}, {self.number=}"


config = load_into(MyConfig, {"my_config": dict(string="initial string", number=0)})

second_config = MyConfig()  # note: no arguments required!
print(f"{config=}\n{second_config=}")
# config=self.string='initial string', self.number=0
# second_config=self.string='initial string', self.number=0

config.update(string="second string", number=1)
# config=self.string='second string', self.number=1
# second_config=self.string='second string', self.number=1



"""
Exposes TypedConfig and load_into for this library.
"""

# SPDX-FileCopyrightText: 2023-present Robin van der Noord <robinvandernoord@gmail.com>
#
# SPDX-License-Identifier: MIT
from .alias import Alias, alias
from .beautify import beautify
from .binary_config import BinaryConfig, BinaryField
from .cls import TypedConfig, TypedMapping, TypedMutableMapping, update
from .core import (
    check_and_convert_data,
    convert_config,
    ensure_types,
    load_data,
    load_into,
    load_into_class,
    load_into_instance,
)
from .dump import asbytes, asdict, asjson, astoml, asyaml
from .helpers import all_annotations, check_type
from .loaders import register_loader as loader
from .postpone import postpone
from .singleton import Singleton, SingletonMeta
from .type_converters import register_converter as converter

__all__ = [
    # beautify,
    "beautify",
    # binary
    "BinaryConfig",
    "BinaryField",
    # cls
    "TypedConfig",
    "TypedMapping",
    "TypedMutableMapping",
    "update",
    # singleton
    "Singleton",
    "SingletonMeta",
    # core
    "check_and_convert_data",
    "convert_config",
    "ensure_types",
    "load_data",
    "load_into",
    "load_into_class",
    "load_into_instance",
    # helpers
    "all_annotations",
    "check_type",
    # postpone
    "postpone",
    # dump
    "asbytes",
    "asdict",
    "astoml",
    "asyaml",
    "asjson",
    # register
    "loader",
    "converter",
    # alias
    "alias",
    "Alias",
]



"""
Register from-to relationship between types, used with load_into(..., convert_types=True).
"""

import types
import typing
from typing import Any

T = typing.TypeVar("T")

Wrapped = typing.Callable[[Any], Any]
Wrapper = typing.Callable[[Wrapped], Wrapped]

CONVERTERS = {}


def register_converter(from_type: type | tuple[type], to_type: type | None | tuple[type | None, ...]) -> Wrapper:
    """
    Register a custom converter that converts from `from_type` to `to_type` using custom logic.

    `from_type` and `to_type` can both be tuples of types,
    but make sure the `to_type`s are compatible with eachother and your method's return value!

    @register_converter(str, bool)
    def str_to_bool(value: str) -> bool:
        ...

    """
    if not isinstance(from_type, tuple):
        from_type = (from_type,)
    if not isinstance(to_type, tuple):
        to_type = (to_type,)

    def wrapper(func: Wrapped) -> Wrapped:
        for _from in from_type:
            for _to in to_type:
                CONVERTERS[(_from, _to)] = func
        return func

    return wrapper


@register_converter(str, bool)
def str_to_bool(value: str) -> bool:
    """
    Used by convert_between, usually for .env loads.

    Example:
        SOME_VALUE=TRUE -> True
        SOME_VALUE=1 -> True
        SOME_VALUE=Yes -> True

        SOME_VALUE  -> None
        SOME_VALUE=NOpe -> False

        SOME_VALUE=Unrelated -> Error
    """
    if not value:
        return False

    first_letter = value[0].lower()
    # yes, true, 1
    if first_letter in {"y", "t", "1"}:
        return True
    elif first_letter in {"n", "f", "0"}:
        return False
    else:
        raise ValueError("Not booly.")


@register_converter(str, (types.NoneType, None))
def str_to_none(value: str) -> typing.Optional[str]:
    """
    Convert a string value of null/none to None, or keep the original string otherwise.
    """
    if value.lower() in {"", "null", "none"}:
        return None
    else:
        return value



import configuraptor


class ConfigWithNones(configuraptor.TypedConfig):
    required: str
    optional: str | None


def test_required():
    config = configuraptor.load_into(ConfigWithNones, {"required": "value", "optional": "value"})

    config.update(required=None)

    assert config.required == "value"

    config.update(required=None, _allow_none=True)
    assert config.required is None


def test_optional():
    config = configuraptor.load_into(ConfigWithNones, {"required": "value", "optional": "value"})

    assert config.optional == "value"

    config.update(optional=None)
    assert config.optional is None

    config.update(optional="new")
    assert config.optional == "new"

    config.update(optional=None, _skip_none=True)
    assert config.optional == "new"



"""
Contains most of the loading logic.
"""

import dataclasses as dc
import io
import os
import typing
import warnings
from pathlib import Path
from typing import Any, Type

import requests

from . import loaders
from .abs import AnyType, C, T, T_data, Type_C
from .alias import Alias, has_alias
from .binary_config import BinaryConfig
from .errors import (
    ConfigErrorCouldNotConvert,
    ConfigErrorInvalidType,
    ConfigErrorMissingKey,
)
from .helpers import (
    all_annotations,
    camel_to_snake,
    check_type,
    dataclass_field,
    find_pyproject_toml,
    is_custom_class,
    is_optional,
    is_parameterized,
)
from .postpone import Postponed
from .type_converters import CONVERTERS


def _data_for_nested_key(key: str, raw: dict[str, typing.Any]) -> dict[str, typing.Any]:
    """
    If a key contains a dot, traverse the raw dict until the right key was found.

    Example:
        key = some.nested.key
        raw = {"some": {"nested": {"key": {"with": "data"}}}}
        -> {"with": "data"}
    """
    parts = key.split(".")
    while parts:
        key = parts.pop(0)
        if key not in raw:
            return {}

        raw = raw[key]

    return raw


def _guess_key(clsname: str) -> str:
    """
    If no key is manually defined for `load_into`, \
    the class' name is converted to snake_case to use as the default key.
    """
    return camel_to_snake(clsname)


def _from_mock_url(url: str) -> str:
    """
    Pytest only: when starting a url with mock:// it is expected to just be json afterwards.
    """
    return url.removeprefix("mock://")


def guess_filetype_for_url(url: str, response: requests.Response = None) -> str:
    """
    Based on the url (which may have an extension) and the requests response \
        (which may have a content-type), try to guess the right filetype (-> loader, e.g. json or yaml).

    Falls back to JSON if none can be found.
    """
    url = url.split("?")[0]
    if url_extension := os.path.splitext(url)[1].lower():
        return url_extension.strip(".")

    if response and (content_type_header := response.headers.get("content-type", "").split(";")[0].strip()):
        content_type = content_type_header.split("/")[-1]
        if content_type != "plain":
            return content_type

    # If both methods fail, default to JSON
    return "json"


def from_url(url: str, _dummy: bool = False) -> tuple[io.BytesIO, str]:
    """
    Load data as bytes into a file-like object and return the file type.

    This can be used by __load_data:
    > loader = loaders.get(filetype)
    > # dev/null exists but always returns b''
    > data = loader(contents, Path("/dev/null"))
    """
    if url.startswith("mock://"):
        data = _from_mock_url(url)
        resp = None
    elif _dummy:
        resp = None
        data = "{}"
    else:
        ssl_verify = os.getenv("SSL_VERIFY", "1") == "1"

        resp = requests.get(url, timeout=10, verify=ssl_verify)
        data = resp.text

    filetype = guess_filetype_for_url(url, resp)
    return io.BytesIO(data.encode()), filetype


def _load_data(
    data: T_data,
    key: str = None,
    classname: str = None,
    lower_keys: bool = False,
    allow_types: tuple[type, ...] = (dict,),
) -> dict[str, typing.Any]:
    """
    Tries to load the right data from a filename/path or dict, based on a manual key or a classname.

    E.g. class Tool will be mapped to key tool.
    It also deals with nested keys (tool.extra -> {"tool": {"extra": ...}}
    """
    if isinstance(data, bytes):
        # instantly return, don't modify
        # bytes as inputs -> bytes as output
        # but since `T_data` is re-used, that's kind of hard to type for mypy.
        return data  # type: ignore

    if isinstance(data, list):
        if not data:
            raise ValueError("Empty list passed!")

        final_data: dict[str, typing.Any] = {}
        for source in data:
            final_data |= load_data(source, key=key, classname=classname, lower_keys=True, allow_types=allow_types)

        return final_data

    if isinstance(data, str):
        if data.startswith(("http://", "https://", "mock://")):
            contents, filetype = from_url(data)

            loader = loaders.get(filetype)
            # dev/null exists but always returns b''
            data = loader(contents, Path("/dev/null"))
        else:
            data = Path(data)

    if isinstance(data, Path):
        with data.open("rb") as f:
            loader = loaders.get(data.suffix or data.name)
            data = loader(f, data.resolve())

    if not data:
        return {}

    if key is None:
        # try to guess key by grabbing the first one or using the class name
        if len(data) == 1:
            key = next(iter(data.keys()))
        elif classname is not None:
            key = _guess_key(classname)

    if key:
        data = _data_for_nested_key(key, data)

    if not data:
        raise ValueError("No data found!")

    if not isinstance(data, allow_types):
        raise ValueError(f"Data should be one of {allow_types} but it is {type(data)}!")

    if lower_keys and isinstance(data, dict):
        data = {k.lower(): v for k, v in data.items()}

    return typing.cast(dict[str, typing.Any], data)


def load_data(
    data: T_data,
    key: str = None,
    classname: str = None,
    lower_keys: bool = False,
    allow_types: tuple[type, ...] = (dict,),
) -> dict[str, typing.Any]:
    """
    Wrapper around __load_data that retries with key="" if anything goes wrong.
    """
    if data is None:
        # try to load pyproject.toml
        data = find_pyproject_toml()

    try:
        return _load_data(data, key, classname, lower_keys=lower_keys, allow_types=allow_types)
    except Exception as e:
        # sourcery skip: remove-unnecessary-else, simplify-empty-collection-comparison, swap-if-else-branches
        # @sourcery: `key != ""` is NOT the same as `not key`
        if key != "":
            return _load_data(data, "", classname, lower_keys=lower_keys, allow_types=allow_types)
        else:  # pragma: no cover
            warnings.warn(f"Data could not be loaded: {e}", source=e)
            # key already was "", just return data!
            # (will probably not happen but fallback)
            return {}


F = typing.TypeVar("F")


def convert_between(from_value: F, from_type: Type[F], to_type: Type[T]) -> T:
    """
    Convert a value between types.
    """
    if converter := CONVERTERS.get((from_type, to_type)):
        return typing.cast(T, converter(from_value))

    # default: just convert type:
    return to_type(from_value)  # type: ignore


def check_and_convert_type(value: Any, _type: Type[T], convert_types: bool, key: str = "variable") -> T:
    """
    Checks if the given value matches the specified type. If it does, the value is returned as is.

    Args:
        value (Any): The value to be checked and potentially converted.
        _type (Type[T]): The expected type for the value.
        convert_types (bool): If True, allows type conversion if the types do not match.
        key (str, optional): The name or key associated with the variable (used in error messages).
                             Defaults to "variable".

    Returns:
        T: The value, potentially converted to the expected type.

    Raises:
        ConfigErrorInvalidType: If the type does not match, and type conversion is not allowed.
        ConfigErrorCouldNotConvert: If type conversion fails.
    """
    if check_type(value, _type):
        # type matches
        return value

    if isinstance(value, Alias):
        if is_optional(_type):
            return typing.cast(T, None)
        else:
            # unresolved alias, error should've already been thrown for parent but lets do it again:
            raise ConfigErrorInvalidType(value.to, value=value, expected_type=_type)

    if not convert_types:
        # type does not match and should not be converted
        raise ConfigErrorInvalidType(key, value=value, expected_type=_type)

    # else: type does not match, try to convert it
    try:
        return convert_between(value, type(value), _type)
    except (TypeError, ValueError) as e:
        raise ConfigErrorCouldNotConvert(type(value), _type, value) from e


def ensure_types(
    data: dict[str, T], annotations: dict[str, type[T]], convert_types: bool = False
) -> dict[str, T | None]:
    """
    Make sure all values in 'data' are in line with the ones stored in 'annotations'.

    If an annotated key in missing from data, it will be filled with None for convenience.

    TODO: python 3.11 exception groups to throw multiple errors at once!
    """
    # custom object to use instead of None, since typing.Optional can be None!
    # cast to T to make mypy happy
    notfound = typing.cast(T, object())

    final: dict[str, T | None] = {}
    for key, _type in annotations.items():
        compare = data.get(key, notfound)
        if compare is notfound:  # pragma: nocover
            warnings.warn("This should not happen since `load_recursive` already fills `data` based on `annotations`")
            # skip!
            continue

        if isinstance(compare, Postponed):
            # don't do anything with this item!
            continue

        if isinstance(compare, Alias):
            related_data = data.get(compare.to, notfound)
            if related_data is not notfound:
                if isinstance(related_data, Postponed):
                    # also continue alias for postponed items
                    continue

                # original key set, update alias
                compare = related_data

        compare = check_and_convert_type(compare, _type, convert_types, key)

        final[key] = compare

    return final


def convert_key(key: str) -> str:
    """
    Replaces '-' and '.' in keys with '_' so it can be mapped to the Config properties.
    """
    return key.replace("-", "_").replace(".", "_")


def convert_config(items: dict[str, T]) -> dict[str, T]:
    """
    Converts the config dict (from toml) or 'overwrites' dict in two ways.

    1. removes any items where the value is None, since in that case the default should be used;
    2. replaces '-' and '.' in keys with '_' so it can be mapped to the Config properties.
    """
    return {convert_key(k): v for k, v in items.items() if v is not None}


def load_recursive(
    cls: AnyType, data: dict[str, T], annotations: dict[str, AnyType], convert_types: bool = False
) -> dict[str, T]:
    """
    For all annotations (recursively gathered from parents with `all_annotations`), \
    try to resolve the tree of annotations.

    Uses `load_into_recurse`, not itself directly.

    Example:
        class First:
            key: str

        class Second:
            other: First

        # step 1
        cls = Second
        data = {"second": {"other": {"key": "anything"}}}
        annotations: {"other": First}

        # step 1.5
        data = {"other": {"key": "anything"}
        annotations: {"other": First}

        # step 2
        cls = First
        data = {"key": "anything"}
        annotations: {"key": str}


    TODO: python 3.11 exception groups to throw multiple errors at once!
    """
    updated = {}

    for _key, _type in annotations.items():
        if _key in data:
            value: typing.Any = data[_key]  # value can change so define it as any instead of T
            if is_parameterized(_type):
                origin = typing.get_origin(_type)
                arguments = typing.get_args(_type)
                if origin is list and arguments and is_custom_class(arguments[0]):
                    subtype = arguments[0]
                    value = [_load_into_recurse(subtype, subvalue, convert_types=convert_types) for subvalue in value]

                elif origin is dict and arguments and is_custom_class(arguments[1]):
                    # e.g. dict[str, Point]
                    subkeytype, subvaluetype = arguments
                    # subkey(type) is not a custom class, so don't try to convert it:
                    value = {
                        subkey: _load_into_recurse(subvaluetype, subvalue, convert_types=convert_types)
                        for subkey, subvalue in value.items()
                    }
                # elif origin is dict:
                # keep data the same
                elif origin is typing.Union and arguments:
                    for arg in arguments:
                        if is_custom_class(arg):
                            value = _load_into_recurse(arg, value, convert_types=convert_types)

                # todo: other parameterized/unions/typing.Optional

            elif is_custom_class(_type):
                # type must be C (custom class) at this point
                value = _load_into_recurse(
                    # make mypy and pycharm happy by telling it _type is of type C...
                    # actually just passing _type as first arg!
                    typing.cast(Type_C[typing.Any], _type),
                    value,
                    convert_types=convert_types,
                )
        elif value := has_alias(cls, _key, data):
            # value updated by alias
            ...
        elif _key in cls.__dict__:
            # property has default, use that instead.
            value = cls.__dict__[_key]
        elif is_optional(_type):
            # type is optional and not found in __dict__ -> default is None
            value = None
        elif dc.is_dataclass(cls) and (field := dataclass_field(cls, _key)) and field.default_factory is not dc.MISSING:
            # could have a default factory
            # todo: do something with field.default?
            value = field.default_factory()
        else:
            raise ConfigErrorMissingKey(_key, cls, _type)

        updated[_key] = value

    return updated


def check_and_convert_data(
    cls: typing.Type[C],
    data: dict[str, typing.Any],
    _except: typing.Iterable[str],
    strict: bool = True,
    convert_types: bool = False,
) -> dict[str, typing.Any]:
    """
    Based on class annotations, this prepares the data for `load_into_recurse`.

    1. convert config-keys to python compatible config_keys
    2. loads custom class type annotations with the same logic (see also `load_recursive`)
    3. ensures the annotated types match the actual types after loading the config file.
    """
    annotations = all_annotations(cls, _except=_except)

    to_load = convert_config(data)
    to_load = load_recursive(cls, to_load, annotations, convert_types=convert_types)

    if strict:
        to_load = ensure_types(to_load, annotations, convert_types=convert_types)

    return to_load


T_init_list = list[typing.Any]
T_init_dict = dict[str, typing.Any]
T_init = tuple[T_init_list, T_init_dict] | T_init_list | T_init_dict | None


@typing.no_type_check  # (mypy doesn't understand 'match' fully yet)
def _split_init(init: T_init) -> tuple[T_init_list, T_init_dict]:
    """
    Accept a tuple, a dict or a list of (arg, kwarg), {kwargs: ...}, [args] respectively and turn them all into a tuple.
    """
    if not init:
        return [], {}

    args: T_init_list = []
    kwargs: T_init_dict = {}
    match init:
        case (args, kwargs):
            return args, kwargs
        case [*args]:
            return args, {}
        case {**kwargs}:
            return [], kwargs
        case _:
            raise ValueError("Init must be either a tuple of list and dict, a list or a dict.")


def _load_into_recurse(
    cls: typing.Type[C],
    data: dict[str, typing.Any] | bytes,
    init: T_init = None,
    strict: bool = True,
    convert_types: bool = False,
) -> C:
    """
    Loads an instance of `cls` filled with `data`.

    Uses `load_recursive` to load any fillable annotated properties (see that method for an example).
    `init` can be used to optionally pass extra __init__ arguments. \
        NOTE: This will overwrite a config key with the same name!
    """
    init_args, init_kwargs = _split_init(init)

    if isinstance(data, bytes) or issubclass(cls, BinaryConfig):
        if not isinstance(data, (bytes, dict)):  # pragma: no cover
            raise NotImplementedError("BinaryConfig can only deal with `bytes` or a dict of bytes as input.")
        elif not issubclass(cls, BinaryConfig):  # pragma: no cover
            raise NotImplementedError("Only BinaryConfig can be used with `bytes` (or a dict of bytes) as input.")

        inst = typing.cast(C, cls._parse_into(data))
    elif dc.is_dataclass(cls):
        to_load = check_and_convert_data(cls, data, init_kwargs.keys(), strict=strict, convert_types=convert_types)
        if init:
            raise ValueError("Init is not allowed for dataclasses!")

        # ensure mypy inst is an instance of the cls type (and not a fictuous `DataclassInstance`)
        inst = typing.cast(C, cls(**to_load))
    elif isinstance(data, cls):
        # already the right type! (e.g. Pathlib)
        inst = typing.cast(C, data)
    else:
        inst = cls(*init_args, **init_kwargs)
        to_load = check_and_convert_data(cls, data, inst.__dict__.keys(), strict=strict, convert_types=convert_types)
        inst.__dict__.update(**to_load)

    return inst


def _load_into_instance(
    inst: C,
    cls: typing.Type[C],
    data: dict[str, typing.Any],
    init: T_init = None,
    strict: bool = True,
    convert_types: bool = False,
) -> C:
    """
    Similar to `load_into_recurse` but uses an existing instance of a class (so after __init__) \
    and thus does not support init.

    """
    if init is not None:
        raise ValueError("Can not init an existing instance!")

    existing_data = inst.__dict__

    to_load = check_and_convert_data(
        cls, data, _except=existing_data.keys(), strict=strict, convert_types=convert_types
    )

    inst.__dict__.update(**to_load)

    return inst


def load_into_class(
    cls: typing.Type[C],
    data: T_data,
    /,
    key: str = None,
    init: T_init = None,
    strict: bool = True,
    lower_keys: bool = False,
    convert_types: bool = False,
) -> C:
    """
    Shortcut for _load_data + load_into_recurse.
    """
    allow_types = (dict, bytes) if issubclass(cls, BinaryConfig) else (dict,)
    to_load = load_data(data, key, cls.__name__, lower_keys=lower_keys, allow_types=allow_types)
    return _load_into_recurse(cls, to_load, init=init, strict=strict, convert_types=convert_types)


def load_into_instance(
    inst: C,
    data: T_data,
    /,
    key: str = None,
    init: T_init = None,
    strict: bool = True,
    lower_keys: bool = False,
    convert_types: bool = False,
) -> C:
    """
    Shortcut for _load_data + load_into_existing.
    """
    cls = inst.__class__
    allow_types = (dict, bytes) if issubclass(cls, BinaryConfig) else (dict,)
    to_load = load_data(data, key, cls.__name__, lower_keys=lower_keys, allow_types=allow_types)
    return _load_into_instance(inst, cls, to_load, init=init, strict=strict, convert_types=convert_types)


def load_into(
    cls: typing.Type[C],
    data: T_data = None,
    /,
    key: str = None,
    init: T_init = None,
    strict: bool = True,
    lower_keys: bool = False,
    convert_types: bool = False,
) -> C:
    """
    Load your config into a class (instance).

    Supports both a class or an instance as first argument, but that's hard to explain to mypy, so officially only
    classes are supported, and if you want to `load_into` an instance, you should use `load_into_instance`.

    Args:
        cls: either a class or an existing instance of that class.
        data: can be a dictionary or a path to a file to load (as pathlib.Path or str)
        key: optional (nested) dictionary key to load data from (e.g. 'tool.su6.specific')
        init: optional data to pass to your cls' __init__ method (only if cls is not an instance already)
        strict: enable type checks or allow anything?
        lower_keys: should the config keys be lowercased? (for .env)
        convert_types: should the types be converted to the annotated type if not yet matching? (for .env)

    """
    if not isinstance(cls, type):
        # would not be supported according to mypy, but you can still load_into(instance)
        return load_into_instance(
            cls, data, key=key, init=init, strict=strict, lower_keys=lower_keys, convert_types=convert_types
        )

    # make mypy and pycharm happy by telling it cls is of type C and not just 'type'
    # _cls = typing.cast(typing.Type[C], cls)
    return load_into_class(
        cls, data, key=key, init=init, strict=strict, lower_keys=lower_keys, convert_types=convert_types
    )
